/* 
 * lbstd2pu-flash.c -- Handle mapping of the flash on Corega LBSTD2PU
 *
 * Base Source is Buffalo BBR-4 By SATO
 *
 * Copyright (C) 2006 SATO, Masuhiro <thomas@fenix.ne.jp>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */
#include <linux/config.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/ioport.h>
#include <linux/kernel.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/mtd/partitions.h>
#include <asm/bootinfo.h>
#include <asm/io.h>
#include <asm/addrspace.h>
#include <asm/adm5120/adm5120.h>


static inline void mpmc_sync(void)
{
	int i;

	for (i = 0; i < 0x1000; i += 4) {
		__raw_readl((void *)(KSEG1 + i));
		asm("sync");
		if  ((MPMC_REG(MPMC_STATUS_REG) & MPMC_WBUF_DIRTY) == 0)
			return;
	}
	printk("write buffer wait count exceeded\n");
}

#define BOOT_FLASH_WINDOW_SIZE	0x0040000
#define BOOT_FLASH_WINDOW_BASE	0x1fc00000
#define USER_FLASH_WINDOW_SIZE	0x00400000
#define USER_FLASH_WINDOW_BASE	0x10000000

#if 0
#define DEBUG_PRINT(fmt, args...)	printk(fmt, ## args);
#else
#define DEBUG_PRINT(fmt, args...)
#endif

static struct mtd_info * lbstd2pu_mtd_init_subr(unsigned long, unsigned long, struct map_info *, struct mtd_partition *, int);
extern int parse_cmdline_partitions(struct mtd_info *, struct mtd_partition **, const char *); 

static struct mtd_info *mtd_info_boot, *mtd_info_user = NULL;


static struct mtd_partition lbstd2pu_boot_partitions[] = {
	{
		name:	"Boot",
		size:	0x00080000,
		offset:	0x00000000,
		mask_flags: MTD_WRITEABLE,
	},
	{
		name:	"Linux Kernel",
		size:	0x00380000,
		offset:	0x00080000,
	},
};

static struct mtd_partition lbstd2pu_user_partitions[] = {
	{
		name:	"User",
		size:	0x01000000,
		offset:	0x00000000,
	},
};


__u8 lbstd2pu_boot_read8(struct map_info *map, unsigned long ofs)
{
	return __raw_readb((void *)map->map_priv_1 + ofs);
}

__u16 lbstd2pu_boot_read16(struct map_info *map, unsigned long ofs)
{
	return __raw_readw((void *)map->map_priv_1 + ofs);
}

__u32 lbstd2pu_boot_read32(struct map_info *map, unsigned long ofs)
{
	return __raw_readl((void *)map->map_priv_1 + ofs);
}

void lbstd2pu_boot_copy_from(struct map_info *map, void *to, unsigned long from, ssize_t len)
{
	memcpy_fromio(to, (void *)map->map_priv_1 + from, len);
}

void lbstd2pu_boot_write8(struct map_info *map, __u8 d, unsigned long adr)
{
	__raw_writeb(d, (void *)map->map_priv_1 + adr);
}

void lbstd2pu_boot_write16(struct map_info *map, __u16 d, unsigned long adr)
{
	__raw_writew(d, (void *)map->map_priv_1 + adr);
}

void lbstd2pu_boot_write32(struct map_info *map, __u32 d, unsigned long adr)
{
	__raw_writel(d, (void *)map->map_priv_1 + adr);
}

void lbstd2pu_boot_copy_to(struct map_info *map, unsigned long to, const void *from, ssize_t len)
{
	memcpy_toio((void *)map->map_priv_1 + to, from, len);
}


static unsigned long map_user_flash(unsigned long adr)
{
	unsigned a22_a23;
	unsigned last_conf0;
	unsigned last_a22_a23;

	last_conf0 = ADM5120_SW_REG(GPIO_conf0_REG);
	last_a22_a23 = last_conf0 >> 30;

	a22_a23 = (adr >> 22) & 0x3;

	if (last_a22_a23 != a22_a23) {
		unsigned tmp;
		ADM5120_SW_REG(GPIO_conf0_REG) = 
			(last_conf0 & 0x3fffffff) | (a22_a23 << 30);
		tmp = ADM5120_SW_REG(GPIO_conf0_REG);
	}

	return adr & 0x003fffff;
}


__u8 lbstd2pu_user_read8(struct map_info *map, unsigned long adr)
{
	__u8 u8_tmp;
	unsigned long flags;

	save_and_cli(flags);
	u8_tmp = __raw_readb((void *)map->map_priv_1 + map_user_flash(adr));
	restore_flags(flags);

	DEBUG_PRINT("lbstd2pu_user_read8(%08x) %02x\n", adr, u8_tmp);

	return u8_tmp;
}

__u16 lbstd2pu_user_read16(struct map_info *map, unsigned long adr)
{
	__u16 u16_tmp;
	unsigned long flags; 

	save_and_cli(flags);
	u16_tmp = __raw_readw((void *)map->map_priv_1 + map_user_flash(adr));
	restore_flags(flags);

	DEBUG_PRINT("lbstd2pu_user_read16(%08x) %04x\n", adr, u16_tmp);

	return u16_tmp;
}

__u32 lbstd2pu_user_read32(struct map_info *map, unsigned long adr)
{
	__u32 u32_tmp;
	unsigned long flags; 

	save_and_cli(flags);
	u32_tmp = __raw_readl((void *)map->map_priv_1 + map_user_flash(adr));
	restore_flags(flags);

	DEBUG_PRINT("lbstd2pu_user_read32(%08x) %08x\n", adr, u32_tmp);

	return u32_tmp;
}

void lbstd2pu_user_copy_from(struct map_info *map, void *to, unsigned long from, ssize_t len)
{
	unsigned long ofs;
	ssize_t n;
	unsigned long flags; 

	DEBUG_PRINT("lbstd2pu_user_copy_from(%08x)\n", from);

	save_and_cli(flags);

	while (len) {
		ofs = map_user_flash(from);
		n = USER_FLASH_WINDOW_SIZE - ofs;
		if (n > len)
			n = len;
		memcpy_fromio(to, (void *)map->map_priv_1 + ofs, n);
		from += n;
		to += n;
		len -= n;
	}

	restore_flags(flags);
}

void lbstd2pu_user_write8(struct map_info *map, __u8 d, unsigned long adr)
{
	unsigned long flags; 

	DEBUG_PRINT("lbstd2pu_user_write8(%02x, %08x)\n", d, adr);

	save_and_cli(flags);
	__raw_writeb(d, (void *)map->map_priv_1 + map_user_flash(adr));
	mpmc_sync();
	restore_flags(flags);
}

void lbstd2pu_user_write16(struct map_info *map, __u16 d, unsigned long adr)
{
	unsigned long flags; 


	DEBUG_PRINT("lbstd2pu_user_write16(%04x, %08x)\n", d, adr);

	save_and_cli(flags);
	__raw_writew(d, (void *)map->map_priv_1 + map_user_flash(adr));
	mpmc_sync();
	restore_flags(flags);
}

void lbstd2pu_user_write32(struct map_info *map, __u32 d, unsigned long adr)
{
	unsigned long flags; 

	DEBUG_PRINT("lbstd2pu_user_write32(%08x, %08x)\n", d, adr);

	save_and_cli(flags);
	__raw_writel(d, (void *)map->map_priv_1 + map_user_flash(adr));
	mpmc_sync();
	restore_flags(flags);
}

void lbstd2pu_user_copy_to(struct map_info *map, unsigned long to, const void *from, ssize_t len)
{
	unsigned long ofs;
	ssize_t n;
	unsigned long flags; 

	DEBUG_PRINT("lbstd2pu_user_copy_to(to=%08x)\n", to);

	save_and_cli(flags);

	while (len) {
		ofs = map_user_flash(to);
		n = USER_FLASH_WINDOW_SIZE - ofs;
		if (n > len)
			n = len;
		memcpy_toio((void *)map->map_priv_1 + ofs, (void *)to, n);
		mpmc_sync();
		from += n;
		to += n;
		len -= n;
	}

	restore_flags(flags);
}


static struct map_info lbstd2pu_boot_map = {
	.name = "lbstd2pu-boot",
	.buswidth = 2,
	.size = 0x00400000,
	.read8 = lbstd2pu_boot_read8,
	.read16 = lbstd2pu_boot_read16,
	.read32 = lbstd2pu_boot_read32,
	.copy_from = lbstd2pu_boot_copy_from,
	.write8 = lbstd2pu_boot_write8,
	.write16 = lbstd2pu_boot_write16,
	.write32 = lbstd2pu_boot_write32,
	.copy_to = lbstd2pu_boot_copy_to
};

static struct map_info lbstd2pu_user_map = {
	.name = "lbstd2pu-user",
	.buswidth = 2,
	.size = 0x01000000,
	.read8 = lbstd2pu_user_read8,
	.read16 = lbstd2pu_user_read16,
	.read32 = lbstd2pu_user_read32,
	.copy_from = lbstd2pu_user_copy_from,
	.write8 = lbstd2pu_user_write8,
	.write16 = lbstd2pu_user_write16,
	.write32 = lbstd2pu_user_write32,
	.copy_to = lbstd2pu_user_copy_to
};


int __init lbstd2pu_mtd_init(void)
{
	mtd_info_boot = lbstd2pu_mtd_init_subr(
		BOOT_FLASH_WINDOW_BASE, BOOT_FLASH_WINDOW_SIZE,
		&lbstd2pu_boot_map, 
		lbstd2pu_boot_partitions,
		ARRAY_SIZE(lbstd2pu_boot_partitions));

	mtd_info_user = lbstd2pu_mtd_init_subr(
		USER_FLASH_WINDOW_BASE, USER_FLASH_WINDOW_SIZE,
		&lbstd2pu_user_map,
		lbstd2pu_user_partitions,
		ARRAY_SIZE(lbstd2pu_user_partitions));

	if (mtd_info_boot == NULL && mtd_info_user == NULL)
		return -1;

	return 0;
}


static struct mtd_info *
lbstd2pu_mtd_init_subr(base, size, map_info, partition, npartitions)
	unsigned long base;
	unsigned long size;
	struct map_info *map_info;
	struct mtd_partition *partition;
	int npartitions;
{
	int nparts;
	struct mtd_info *mtd_info = NULL;
	struct mtd_partition *parts = NULL;

	if (!request_mem_region(base, size, map_info->name))
		return NULL;

	map_info->map_priv_1 = (unsigned long)ioremap(base, size);
	if (!map_info->map_priv_1) {
		printk(KERN_NOTICE "Failed to ioremap %s\n", map_info->name);
		release_mem_region(base, size);
		return NULL;
	}

	mtd_info = do_map_probe("cfi_probe", map_info);
	if (!mtd_info) {
		iounmap((void *)map_info->map_priv_1);
		map_info->map_priv_1 = 0;
		release_mem_region(base, size);

		return NULL;
	}

	mtd_info->module = THIS_MODULE;

#ifdef CONFIG_MTD_CMDLINE_PARTS
	nparts = parse_cmdline_partitions(mtd_info, &parts, map_info->name);
	if (nparts <= 0)
		parts = NULL;
#endif
	if (parts == NULL) {
		parts = partition;
		nparts = npartitions;
	}
		
	add_mtd_partitions(mtd_info, parts, nparts);

	return mtd_info;
}



static void __exit lbstd2pu_mtd_cleanup(void)
{
	if (mtd_info_boot) {
		del_mtd_partitions(mtd_info_boot);
		map_destroy(mtd_info_boot);
		iounmap((void *)lbstd2pu_boot_map.map_priv_1);
		lbstd2pu_boot_map.map_priv_1 = 0;
		release_mem_region(BOOT_FLASH_WINDOW_BASE,
				BOOT_FLASH_WINDOW_SIZE);
		mtd_info_boot = NULL;
	}
	if (mtd_info_user) {
		del_mtd_partitions(mtd_info_user);
		map_destroy(mtd_info_user);
		iounmap((void *)lbstd2pu_user_map.map_priv_1);
		lbstd2pu_user_map.map_priv_1 = 0;
		release_mem_region(USER_FLASH_WINDOW_BASE,
				USER_FLASH_WINDOW_SIZE);
		mtd_info_user = NULL;
	}
}


module_init(lbstd2pu_mtd_init);
module_exit(lbstd2pu_mtd_cleanup);

MODULE_LICENSE("GPL");

